// EndUnit.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The EndUnit class instances represent virtual or actuale 
 * places or warehouse were temporary finished sub-recipes (produced internally
 * or obtained as procurements) wait to be used in other recipes via a "p"
 * (procurement) step<br><br>
 *
 * an endUnit has its internal number equal to the concluding code
 * used to finish/identify a temporary finished sub-recipe
 *
 * @author Pietro Terna
 */
public class EndUnit extends SwarmObjectImpl implements PTHistogramPlottable
{

    /** the number identifying the unit which is also its end code */
    public int unitNumber;
    /** the list containing order temporary kept in this virtual (or actual)
     *  place */
    public ListImpl  temporaryList;
    /** the modelSwarm (to which we can apply the 
     * setProductionTimeAndRecipeLength(...) method, if any */
    VEFrameModelSwarm vEFrameModelSwarm;


    /**
     * the constructor for EndUnit
     */
    public EndUnit (Zone aZone, VEFrameModelSwarm mo)
    {
	// Call the constructor for the Unit parent class.
	super(aZone);

	// the model (for future use)
	vEFrameModelSwarm = mo;

	// creating the internal list temporary the items temporary in hold
	// hold
	temporaryList = new ListImpl (getZone());
    }

    /** set unit number */
    public void setUnitNumber(int un)
    {
	// the unit id number.
	unitNumber=un;
	// announce the unit to the console
	if(StartVEFrame.verbose)
	System.out.println("End unit number " + unitNumber + 
			   " has been created.");
     }

    /** get unit number */
    public int getUnitNumber()
    {
	// the unit id number.
	return unitNumber;
    }


    /**
     * Putting an order in the temporaryList of the unit
     */
    public void setTemporaryList(Order anOrder)
    {
        temporaryList.addLast(anOrder);
	if(StartVEFrame.verbose)
	    System.out.println("The end unit " + unitNumber +
			       " has received the order # "
			       + anOrder.getOrderNumber());
    }

    /**
     * Removing an order from the temporaryList of the unit
     * and returning its address
     */
    public Object removeFromTemporaryList()
    {
	Order removed;
        
	removed = (Order) temporaryList.removeFirst();
	if(StartVEFrame.verbose)
	    System.out.println("The end unit " + unitNumber +
			       " has removed the order # "
			       + removed.getOrderNumber());
	return removed;
    }

    /**
     * return the temporary list length
     */
    public int getTemporaryListLength()
    {
	return temporaryList.getCount();
    }


    /* PTHistogramPlottable interface method: getLabel() */
    public String getLabel(){
	return "" + getUnitNumber();
    }

    /* PTHistogramPlottable interface method: getValueToPlot() */
    public double getValueToPlot(){
	return (double) getTemporaryListLength();
    }

}
