// InventoryRuleMaster.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The InventoryRuleMaster is used to make decisions about inventories.
 *
 * @author Marco Remondino and Pietro Terna, with Paolo Pelligra
 */
public class InventoryRuleMaster extends SwarmObjectImpl {

    /** the number of the unit which is calling us */
    public int unitNumber;
    /** the max quantity to be kept in the warehouse (only integer
	values at present) */
    public int maxInWarehouses;
    /** the min quantity to be kept in the warehouse (only integer
	values at present) */
    public int minInWarehouses;
    /** the warehouse of the unit which is calling us */
    public Warehouse theUnitWarehouse;
    /** the number of newses necessary to decide the productio of 
     *  inventories */
    public int nOfNewsesToProduce;
    /** the number of newses to be cleared after the decision of producing
     *  to increase inventories (a simple solution is to have it equal to 
     *  nOfNewsesToProduce) */
    public int nOfNewsesToBeCleared;

    /** the news list */
    public ListImpl newsList;
    /** a news not yet cleared by the InventoryRuleMaster */
    public News toBeClearedNews;

    /**
     * constructor for InventoryRuleMaster
     */
    public InventoryRuleMaster (Zone aZone, int max, int min,
				int nop, int noc) 
    {
	// call the constructor for the warehouse parent class
	super(aZone);

	// the min/max quantity to be kept in a warehouse
	maxInWarehouses=max;
        minInWarehouses=min;
	nOfNewsesToProduce=nop;
	nOfNewsesToBeCleared=noc;

    }

    /** the unit which is calling us is idle and it is asking us if it has to
	produce or not, to increase inventories */
    public boolean increaseInventories(int un, Warehouse aw) 
    {
	unitNumber=un;
	theUnitWarehouse=aw;
	if(StartVEFrame.verbose)
	    System.out.println("* Unit #" + unitNumber +
			       " is calling the inventoryRuleMaster");
	if(StartVEFrame.verbose)
	    System.out.println("* Warehouse #" +
			       theUnitWarehouse.getWarehouseNumber() +
			       " contains " +
			       theUnitWarehouse.getInventoryQuantity() +
			       " inventory units");

	if (theUnitWarehouse.getInventoryQuantity()<maxInWarehouses)
	    {
		if(StartVEFrame.verbose)
		    System.out.println("* Unit #" + unitNumber +
				       " has to increase inventories");
		return true;
	    }
	else if(StartVEFrame.verbose)
	    System.out.println("* The warehouse of unit #" +
			       unitNumber + " is already full.");

	return false;
    }

    /** the unit which is calling us uses newses, is idle and it is asking 
     * us if it has to produce or not to increase inventories; we check the 
     * newses not yet cleared and we clear them
     * we agree to the production of inventories if the warehouse level is 
     * < maxInWarehouses && (we have a news signalling and incoming 
     * production step || the warehouse level is < minInWarehouses)
     */

    public boolean increaseInventoriesUsingNewses
	(int un, Warehouse aw, ListImpl nl) 
    {
        int i;
        int count;
	int countCleared;
        newsList= nl;
        unitNumber=un;
	theUnitWarehouse=aw;

        if(StartVEFrame.verbose)
	    System.out.println("* Unit #" + unitNumber +
			" is calling the inventoryRuleMaster using newses");
	if(StartVEFrame.verbose)
	    System.out.println("* Warehouse #" +
			       theUnitWarehouse.getWarehouseNumber() +
			       " contains " +
			       theUnitWarehouse.getInventoryQuantity() +
			       " inventory units");

	// check the newsList for newses not yet cleared, agree to increase 
	// the inventory (if it is possible) and clear the newses

	count=0; countCleared=0;
	for (i=0;i<newsList.getCount();i++)
	    {
		toBeClearedNews = (News) newsList.atOffset(i);
		if(toBeClearedNews.getNotCleared())
		    {
			count++; countCleared++;
			if(countCleared<=nOfNewsesToBeCleared)
			    toBeClearedNews.setClearedByInventoryRuleMaster();
			if(StartVEFrame.verbose)
			    System.out.println("the InventoryRuleMaster "
					       + "is clearing a news in unit " 
					       + unitNumber);
		    }
	    }

       if(count>=nOfNewsesToProduce && 
	  theUnitWarehouse.getInventoryQuantity()<maxInWarehouses)
	   {
	       if(StartVEFrame.verbose)
		   System.out.println("* Unit #" + unitNumber +
				      " is increasing inventories via newses");
	       return true;
	   }
       else if(count<nOfNewsesToProduce && 
	       theUnitWarehouse.getInventoryQuantity()<minInWarehouses)
	   {
	       if(StartVEFrame.verbose)
		   System.out.println("* Unit #" + unitNumber +
				   " is increasing inventories via minInWarehouses");
	       return true;
	   }

       else
	   {
	       if(StartVEFrame.verbose)
		   System.out.println("* The warehouse of unit #" + unitNumber + 
				      " is already full or newsList is empty" +
				      " and we are over minInWarehouses quantity");
                         return false;
                         }
      }
}






