// Order.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The form containing our Order metaphoric object
 *
 * @author Pietro Terna (with Paolo Pelligra and Dario Landini contibutions)
 */
public class Order  extends SwarmObjectImpl{

    /** the number of steps in the product recipe */
    public int stepNumber;
    /** the number of this order */
    public int orderNumber;
    /** the day in which the order as been generated */
    public int startingDay;
    /** integer vector, containing the recipe steps */
    public int[] orderRecipe;
    /** a temporary vector to redefine the dimension of orderRecipe
     *  if necessary (using  "p" or "c" steps) */ 
    public int[] orderRecipeTmp;
    /** integer vector, recording the performed phases */
    public int[] orderState;
    /** the sequential number of the recipe referring to which
     * a news will be sent */
    public int numberOfTheStepForWhichANewsIsSent;
    /** the fixed and variable costs of the order */
    public float cumulatedFixedCost, cumulatedVariableCost;
    /** the sum of variable and fixed costs of the order */
    public float orderTotalCosts = 0;

    /** a procurement specication set */
    public ProcurementSpecificationSet aProcurementSpecificationSet;
    /** the list of procurement specications */
    public ListImpl procurementSpecificationSetList;

    /** the list of procured items */
    public ListImpl procuredItemList;

    /** the model */
    VEFrameModelSwarm vEFrameModelSwarm;

  /**
   * constructor for Order
   */
  public Order (Zone aZone, int n, int sd, int sn, int [] r, 
		VEFrameModelSwarm model) {

     // Call the constructor for the Order parent class
     super(aZone);

     orderNumber=n;
     stepNumber=sn; // it contains also the length of "p" or "c" sequences
     startingDay=sd;
     orderRecipe= new int[stepNumber];
     orderState = new int[stepNumber];
     // the model
     vEFrameModelSwarm = model;

     int i, ii, j, jj;
     procurementSpecificationSetList = new ListImpl(getZone());
     procuredItemList = new ListImpl(getZone());

     i=0; j=0;
     // copying the recipe, analysing and using "p" o "c" sequences
     while (j<stepNumber)
	 {
	     if (r[j]==0){
		 System.out.
		     println("Internal error in order generation: " + 
			     "look at the constructor of Order class.");
		 System.exit(1);
	     }

	     if (r[j]>0) orderRecipe[i]=r[j];

	     // "p" or "c" case
	     if (r[j]<0)
		 if(r[j]==-1){
		     jj=j;
		     j += r[j+1] + 2;
		     if (! (j<stepNumber)){System.out.
			    println(
			    "Internal dimension error analysing the order: " + 
			    "look at the constructor of Order class.");
		     System.exit(1);
		     }
 
		     // making a procurementSpecificationSet object
		     aProcurementSpecificationSet = new 
			 ProcurementSpecificationSet(getZone(), 
						     orderNumber, i);

		     procurementSpecificationSetList.
			 addLast(aProcurementSpecificationSet);

		     // setting specifications into 
		     // aProcurementSpecificazioneSet
		     aProcurementSpecificationSet.
			 setSpecificationSet(jj,r);

		     // copying the recipe
		     orderRecipe[i] = -r[j];
		 }
		 else{System.out.
			  println("Internal error analysing the order: " + 
				  "look at the constructor of Order class.");
		 System.exit(1);}

	     i++; j++;

	 }

     // i is now the internal length of vectors, so we redefine the 
     // stepNumber
     stepNumber=i;;

     // redefine orderRecipe vector
     orderRecipeTmp = new int[stepNumber];
     System.arraycopy(orderRecipe,0,orderRecipeTmp,0,stepNumber);
     orderRecipe=orderRecipeTmp;

     // announcing the order presence to the console
     if(StartVEFrame.verbose)
	 System.out.println("New order (#" + orderNumber + ")");

     // setting the state vector (0=empty; 1=done)
     if(StartVEFrame.verbose)System.out.print("orderState  vector ");
     for (i=1;i<=stepNumber;i++) {
	 orderState[i-1]=0;
	 if(StartVEFrame.verbose)System.out.print(orderState[i-1] + " ");
     }
     if(StartVEFrame.verbose)System.out.println();

     // announcing the recipe vector
     if(StartVEFrame.verbose)System.out.print("orderRecipe vector ");
     for (i=1;i<=stepNumber;i++)
	 if(StartVEFrame.verbose)System.out.print(orderRecipe[i-1] + " ");
     if(StartVEFrame.verbose)System.out.println();
  }

  /**
   * production next step in Order (it is in the orderRecipe vector,
   * at the same index value in which we find the first 0 in orderState
   * vector)
   */
    public int getNextStep () {

	int i, nextStep;

	nextStep=0;
	for (i=0;i< stepNumber && nextStep==0;i++)
	    if (orderState[i]==0)nextStep=getOrderRecipe(i);

	// if no 0 position can be found the order is finished
	// in this case we return 0

	return nextStep;
  }

  /**
   * sign of the production next step in Order
   */
    public boolean getNextStepNegative () {

	int i, nextStep; 
	boolean negative;

	negative=false;
	nextStep=0;
	for (i=0;i< stepNumber && nextStep==0;i++)
	    if (orderState[i]==0)nextStep=orderRecipe[i];

	if(nextStep<0)negative=true;
	return negative;
  }

  /**
   * change the sign of the production next step in Order, if negative
   */
    public void setNextStepToPositive () {

	int i, nextStep; 

	nextStep=0;
	for (i=0;i< stepNumber && nextStep==0;i++){
	    if (orderState[i]==0)nextStep=orderRecipe[i];
	    if(nextStep<0)orderRecipe[i] *= -1.;
	}
  }

  /**
   * Done step in Order
   */
    public void setDoneStep () {

	int i, done;

	done=0;
	for (i=0;i<= stepNumber && done==0;i++)
	    if (orderState[i]==0)
		{orderState[i]=1;
		done=1;
		}
    }

  /**
   * returning the orderNumber of this order
   */
    public int getOrderNumber () {
	return orderNumber;
  }

  /**
   * returning the stepNumber
   */
    public int getStepNumber () {

	return stepNumber;
  }

  /**
   * returning the startingDay
   */
    public int getStartingDay () {

	return startingDay;
  }

  /**
   * returning how many steps remain to terminate the production
   */
    public int getHowManyStepsToBeDone () {

	int i, stepsToBeDone;
	for (i=0;i<stepNumber && orderState[i]==1;i++); // to find the first position
	                                                // where orderState[i] != 1

	stepsToBeDone=stepNumber-i;
   
	if(StartVEFrame.verbose)
	    System.out.println
	       ("stepsToBeDone  " + stepsToBeDone+ " in order #"+ orderNumber);

	return stepsToBeDone;
  }

  /**
   * returning the production phase at a specific step
   */

    public int getProductionToBeDoneAtStepAheadNumber (int sn) {

	int i;
	int step;

	step = sn;
	for (i=0;i< stepNumber && orderState[i]==1;i++); // to find the first position
	                                                 // where orderState[i] != 1

	numberOfTheStepForWhichANewsIsSent=i+step-1;

	return getOrderRecipe(numberOfTheStepForWhichANewsIsSent);
    }

 /**
   * returning the sequential number of the recipe referring to which
   * a news will be sent
   */
    public int getNumberOfTheStepForWhichANewsIsSent () {

	return numberOfTheStepForWhichANewsIsSent;
  }

  /** returning the value of the production already done */
    public float getOrderEvaluation(){
	int globalOrderState = 0;
	float value = 0;

	// evaluate the number of steps done
	for(int i=0; i<orderState.length; i++){
	    globalOrderState += orderState[i];
	}

	value = globalOrderState * 
	    vEFrameModelSwarm.getRevenuePerEachRecipeStep();
	return value;
    }

  /** setting fixed and variable costs for the orders in production
   *  or finished */
    public void setFixedAndVariableOrderCosts(float fc, float vc){

	cumulatedFixedCost += fc;
	cumulatedVariableCost += vc;
	orderTotalCosts = (cumulatedFixedCost + cumulatedVariableCost);
  }

  /** getting the total of costs of the order */ 
    public float getOrderTotalCosts(){
	return orderTotalCosts;
    }

    /** getting the content of orderRecipe[i], with the sign changed if
     * it is < 0 */
    public int getOrderRecipe(int i){
	if(orderRecipe[i]>=0) return orderRecipe[i];
	else                  return - orderRecipe[i];
    }

    /** getting the procurementSpecificationSet related to the nextStep
     *  (the current one, if the order is in a waiting list for next step
     *  to be accomplished<br>
     *  if no procurementSpecificationSet exists for the next step,
     *  a null value is returned */
    public Object getPendingProcurementSpecificationSet()
    {
	int i, position;
	ProcurementSpecificationSet pendingProcurementSpecificationSet;

	position=-1;
	for (i=0;i< stepNumber && position==-1;i++)
	    if (orderState[i]==0)position=i;

	pendingProcurementSpecificationSet = null;

	for(i=0;i<procurementSpecificationSetList.getCount();i++){
	    if(position == ((ProcurementSpecificationSet)
			    procurementSpecificationSetList.
			    atOffset(i)).
	       getPositionInRecipe())
		pendingProcurementSpecificationSet =
		    (ProcurementSpecificationSet)
		    procurementSpecificationSetList.
		    atOffset(i);
	}

	return pendingProcurementSpecificationSet;
    }

    /** adding an item to the procuredItemList */
    public void setProcuredItemList(Order o){
	procuredItemList.addLast(o);
    }

    /** drop the order */
    public void drop(){
	Order aProcuredItem;

	// dropping our procurementSpecificationSet
	while(procurementSpecificationSetList.getCount()>0)
	    ((ProcurementSpecificationSet) 
	     procurementSpecificationSetList.removeFirst()).drop();

	// dropping procured items (which in turn will be dropping their
	// procurementSpecificationSet and their procured items
	while(procuredItemList.getCount()>0)
	    {
		aProcuredItem = (Order) procuredItemList.removeFirst();
		((ListImpl) vEFrameModelSwarm.getOrderList()).
		    remove(aProcuredItem); 
		aProcuredItem.drop();
	    }

	super.drop();
}

}
