// OrderGenerator.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The OrderGenerator is used at present to simulate the market
 * behavior.
 *
 * @author Pietro Terna */
public class OrderGenerator  extends SwarmObjectImpl{

    /** How many unit we have, used at present to define the max value
	used to generate the 'digits' composing our recipes */

    /** the max length of a recipe */
    public int maxStepNumber;

    /** the max length of a step in a recipe */
    public int maxStepLength;

    /** used to record the number of generated orders*/
    public int orderCount;

    /** a specific order*/
    public Order anOrder;
    /** the list containig the operating units*/
    public ListImpl unitList;
   /** the list containig the end units*/
    public ListImpl endUnitList;
    /** the list containig all the orders*/
    public ListImpl orderList;
    /** the model (to which we apply mark up of the order revenue */
    VEFrameModelSwarm vEFrameModelSwarm;

    /** the integer vector containing a specific production recipe */
    public int[] orderRecipe1; 

    /** the integer vector containing the unit of time spent for each step 
     *  in the recipe; il will determine the tranformation of the orderRecipe1 
     *  vector into the orderRecipe2 vector containing the repetitions
     *  of the steps, to reproduce the length of each step in time
     *
     *  NB this formalism is internal to internal (in external formalism
     *  we have 'step s/h/d time'
     *
     *  as an examples
     * 
     *  orderRecipe1      contains  1 12  7  3
     *  stepLengthInOrder contains  1  3  2  2
     *
     *  orderRecipe2      contains  1 12 12 12  7  7  3  3 
     */
    public int[] stepLengthInOrder;

    /** the integer vector containing a specific production recipe,
     *  developed following the length of each step
     */
    public int[] orderRecipe2;

    /** a temporary vector to redefine the dimension of orderRecipe2
     *  if necessary (using  "p" or "c" steps) */ 
    public int[] orderRecipeTmp;

    /** the dictionary to be used to choose the steps to be included in 
        a recipe */
    public int[] dictionary;
    /** length of the dictionary */
    public int dictionaryLength;

    // units
    Unit aUnit;
    EndUnit anEndUnit;


    /**
     * the constructor for OrderGenerator
     */
    public OrderGenerator (Zone aZone, int msn, int msl, ListImpl ul, 
			   ListImpl eul, ListImpl ol, 
			   VEFrameModelSwarm mo) 
    {

	// Call the constructor for the OrderGenerator parent class.
	super(aZone);

	orderCount=0;
	maxStepNumber=msn;
	maxStepLength=msl;
	unitList=ul;
	endUnitList=eul;
        orderList=ol;

	orderRecipe1 = new int[maxStepNumber];
	stepLengthInOrder = new int[maxStepNumber];
	orderRecipe2 = new int[maxStepNumber*maxStepLength];

        vEFrameModelSwarm = mo;
    }

    /** building the dictionary containing the steps to be included in the
	recipes */
    public void setDictionary()
    {
	int i, ii, iii;
	// how many words in the dictionary?

	// starting point: the no. of endUnits 
	dictionaryLength=endUnitList.getCount(); 

	for (i=0;i<unitList.getCount();i++)
	    {
		aUnit=(Unit) unitList.atOffset(i);
		dictionaryLength += aUnit.getNOfPhasesToDealWith();
	    }   
	if(StartVEFrame.verbose)
	    System.out.println(
	 "Length of the dictionary (# of codes identifying steps in recipes): "
	 + dictionaryLength);

	//creating the dictionary
	dictionary = new int[dictionaryLength];

	// loading terms into the dictionary (from endUnits and then units)
	i=0;
	for (ii=0;ii<endUnitList.getCount();ii++)
	    {
		anEndUnit=(EndUnit) endUnitList.atOffset(ii);
		dictionary[i++]=anEndUnit.getUnitNumber();
	    }
	for (ii=0;ii<unitList.getCount();ii++)
	    {
		aUnit=(Unit) unitList.atOffset(ii);
		for (iii=0;iii<aUnit.getNOfPhasesToDealWith();iii++)
		    dictionary[i++]=aUnit.getProductionPhase(iii);
	    }

	if(StartVEFrame.verbose)
	    for (i=0;i<dictionaryLength;i++)
		System.out.println(i + "  " + dictionary[i]);
    }


    /** creating an order form, choosing its recipe length and content */
    public void createRandomOrderWithNSteps() {
	int i, ii, iii, k, count, 
	    randomStepNumber, randomStepNumber0, randomStep;
	boolean unitNotFound, finished;

	orderCount++;

	randomStepNumber=Globals.env.uniformIntRand.getIntegerWithMin$withMax
		    (1, maxStepNumber);

	// generating the sequence in the recipe

	// putting 0 in the vector may be useful for future software 
	// modifications
	for (i=0;i<maxStepNumber;i++)orderRecipe1[i]=0;
	for (i=0;i<maxStepNumber;i++)stepLengthInOrder[i]=0;

	finished=false;
	randomStepNumber0=0;

	for (i=0;i<randomStepNumber && ! finished;i++)
	    {
		randomStepNumber0=i+1;

		// generating a random step and stating the probability
		// of concluding the recipe with an "e" end code
		if(0.10 >= Globals.env.uniformDblRand.
		   getDoubleWithMin$withMax(0,1) && 
		   endUnitList.getCount()>0)		
		    randomStep = Globals.env.uniformIntRand.
			getIntegerWithMin$withMax(0, endUnitList.getCount()-1);
		else
		    randomStep = Globals.env.uniformIntRand.
			getIntegerWithMin$withMax(endUnitList.getCount(),
						  dictionaryLength-1);
		orderRecipe1[i]= dictionary[randomStep];

		// check if an end code has been generated
		if (randomStep < endUnitList.getCount())finished=true;

		// generating the length of each step
		if (maxStepLength>1 && ! finished) // end step has lenght 1
		    stepLengthInOrder[i]=
			Globals.env.uniformIntRand.getIntegerWithMin$withMax
			(1, maxStepLength);
		else 
		    stepLengthInOrder[i]=1;
	    }

	// putting (i) sequence and (ii) length, together
	count=0;
	for (i=0;i<randomStepNumber0;i++)
	    for (ii=0;ii<stepLengthInOrder[i];ii++)
		{

		    // introducing "p" or "c" sequences

		    // the "p" strucuture is
		    // p k a1 a2 .... ak
		    // where k is the # of items to be procured and 
		    // a1, a2, ..., ak are the codes of those items 
		    // (corresponding to "e" or endUnit codes in sub-recipes
		    // related to internal or external sub-processes

		    // the recipe scheme is
		    // external recipe (with the time specifications and 
		    // a tick corresponding to 1 sec. or to an analogous
		    // minimum fraction of the time
		    
		    // 11 s 3 p 3 101 102 103 22 s 2
		    
		    // internally, in orderGenerator, with p reported as -1

		    // 11 11 11 -1 3 101 102 103 22 22

		    // into anOrder

		    // 11 11 11 -22 22
		    // with a procurementSpecifications object having the same 
		    // # of the order and a position paramer relative to the
		    // recipe sequence (here position = 4)

		    // and, when anOrder comes back from procurementAssembler,
		    // 11 11 11 22 22
		    // always with the procumerementSpecifications object
		    
		    if(endUnitList.getCount() > 0)
			{
			    if (ii==0 && 0.05 >= Globals.env.uniformDblRand.
				getDoubleWithMin$withMax(0,1)){

				// length of the "p" or "c" step
				k=Globals.env.uniformIntRand.
				    getIntegerWithMin$withMax
				    (1, endUnitList.getCount());

				// increase the length of orderRecipe2 vector
				orderRecipeTmp = new int[orderRecipe2.
							length+2+k];
				System.arraycopy(orderRecipe2,0,
						 orderRecipeTmp,0,
						 orderRecipe2.length);
				orderRecipe2=orderRecipeTmp;

				//the code of process p
				orderRecipe2[count++]= -1; 

				// the length
				orderRecipe2[count++]=k;   
				
				// k codes, chosen randomly
				for (iii=0;iii<k;iii++)   
				    orderRecipe2[count++]=
					((EndUnit) endUnitList.
					 atOffset(
						  Globals.env.uniformIntRand.
						  getIntegerWithMin$withMax
						  (0, endUnitList.getCount()-1)
						  )).getUnitNumber();
			    }
			}

		    // copying normal steps
		    orderRecipe2[count++]=orderRecipe1[i];

		}
	
	if(StartVEFrame.verbose)
	{
	    for (iii=0; iii<count;iii++)
		System.out.print(orderRecipe2[iii]+ " ");
	    System.out.println(" internal recipe for order " + orderCount);
	}


	// creating an order
        // vector dimension in anOrder is count: count is a step ahead (see 
        // last count++ operation) but starting with 0 is the correct number
        // of elements
	anOrder = new Order(getZone(), orderCount,
			    Globals.env.getCurrentTime(),
			    count, orderRecipe2, 
			    vEFrameModelSwarm);

	if(StartVEFrame.verbose)
	{
	    for (iii=0; iii<anOrder.orderRecipe.length;iii++)
		System.out.print(anOrder.orderRecipe[iii]+ " ");
	    System.out.println(" internal recipe into the order " +
			       orderCount);
	}
	

        // add the active orders to the general order list (they will be
	// eliminated when dropped in a unit, being finished); this
	// list has been introduced for accounting purposes [may be it would
	// be better subtitute it with a get to the units to know their
	// waiting lists] 
        orderList.addLast(anOrder);

	/**
	 * sending the order to the first production unit
	 * (we are acting as the Front End of the VE)
	 */
	unitNotFound=true;
	for (i=0;i<unitList.getCount() && unitNotFound;i++)
	    {
		aUnit=(Unit) unitList.atOffset(i);
		for (iii=0;iii<aUnit.getNOfPhasesToDealWith() &&  
			 unitNotFound;iii++)
		    {
		       if(anOrder.getNextStep()==aUnit.getProductionPhase(iii))
			    {
				unitNotFound=false;
				aUnit.setWaitingList(anOrder);
			    }
		    }
	    }

	// considering also endUnits
	for (i=0;i<endUnitList.getCount() && unitNotFound;i++)
	    {
		anEndUnit=(EndUnit) endUnitList.atOffset(i);
		if(anOrder.getNextStep()==anEndUnit.getUnitNumber())
		    {
			unitNotFound=false;
			anEndUnit.setTemporaryList(anOrder);
		    }
	    }



	if(unitNotFound){
	    System.out.println("OrderGenerator internal error: receiving "+
			       "unit does not exixts");
	    System.exit(1);
	}

    }

}
