//  ProcurementAssembler.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The ProcurementAssembler class instances are unit 
 * assembling the elements of a procurement process
 *
 * @author Pietro Terna 
 */
public class ProcurementAssembler extends SwarmObjectImpl implements 
    PTHistogramPlottable{


    /** the unit we are assembling for */
    Unit myUnit;
    /** the list of the order to be assembled with the procurements */
    public ListImpl waitingList;
    /** the list of end units */
    public ListImpl endUnitList;
    /** the vector of the quantities in the end units */
    public int [] quantitiesInEndUnits;
    /** the vector of the labels fo the end units */
    public int [] endUnitLabels;
    /** a procurement specification set */
    ProcurementSpecificationSet pendingProcurementSpecificationSet;

    /**
     * the constructor for ProcurementAssembler
     */
    public ProcurementAssembler (Zone aZone, ListImpl eul)
    {
	// Call the constructor for the parent class.
	super(aZone);
	int i;

	// the end unit list
	endUnitList = eul;
	quantitiesInEndUnits = new int[endUnitList.getCount()];
	endUnitLabels = new int [endUnitList.getCount()];

	for (i=0; i<endUnitList.getCount(); i++)
	    endUnitLabels[i]= ((EndUnit) endUnitList.atOffset(i)).
		getUnitNumber();
    
	// the list procurement processes to be assembled
	waitingList = new ListImpl (getZone());

    }

    /** setting the unit we are assembling for */
    public void setUnit(Unit u){
	myUnit = u;
    }

    /** adding an order to the waitingList and eliminating the negative sign
     *  in step */
    public boolean setProcurementWaitingList(Order anOrder){
	waitingList.addLast(anOrder);
	anOrder.setNextStepToPositive();
	return true;
    }

    /** verifying procurements and sending orders to our unit */
    public void checkProcurementsAndSendBackOrders(){
	int n, i, ii, nmax;
	boolean completed;
	Order anOrder;

	nmax = waitingList.getCount();
	if(nmax>0)
	    for (n=0;n<nmax;n++)
		{
	if(StartVEFrame.verbose)
		    System.out.println(" ");
		    // create the table of the quantities into the end units
		    for (i=0; i<endUnitList.getCount();i++){
			quantitiesInEndUnits[i]=
			    ((EndUnit) endUnitList.atOffset(i)).
			    getTemporaryListLength();
	if(StartVEFrame.verbose)
			System.out.print("eu " + endUnitLabels[i] + " " +
					 quantitiesInEndUnits[i] + "   ");
		    }
	if(StartVEFrame.verbose)
		    System.out.println(
				   " as seen in procurementAssembler of unit "
				   + myUnit.getUnitNumber());

		    anOrder = (Order) waitingList.removeFirst();
		    pendingProcurementSpecificationSet = 
			(ProcurementSpecificationSet)
			 anOrder.getPendingProcurementSpecificationSet();

		    if(pendingProcurementSpecificationSet == null){
			System.out.println("Internal error in" +
				       " procurementAssembler of unit " +
				       myUnit.getUnitNumber() +
				       " - anOrder waiting for procurement "+
				       "has no procurement specifications.");
			System.exit(1);
		    }

	if(StartVEFrame.verbose)
		    {
		    System.out.print("proc. spec. of order " +
				       pendingProcurementSpecificationSet.
				       getNumber() + " at pos. " +
				       pendingProcurementSpecificationSet.
				       getPositionInRecipe() + " items ");
		    for (i=0;i<pendingProcurementSpecificationSet.
			     getNumberOfItemsToBeProcured();i++)
			System.out.print(pendingProcurementSpecificationSet.
					 getItemToBeProcured(i) + " ");
		    System.out.println(" ");
			}

		    // checking endUnits for the required procurements
		    for (i=0;i<pendingProcurementSpecificationSet.
			     getNumberOfItemsToBeProcured();i++)
			for (ii=0; ii<endUnitList.getCount(); ii++)
			    if(endUnitLabels[ii]==
			       pendingProcurementSpecificationSet.
			       getItemToBeProcured(i))
				quantitiesInEndUnits[ii]--;
	if(StartVEFrame.verbose)
		    for(ii=0; ii<endUnitList.getCount(); ii++)
			System.out.print("eu " + endUnitLabels[ii] + " " +
					 quantitiesInEndUnits[ii] + "   ");
	if(StartVEFrame.verbose)
		    System.out.println(
				" as residual in procurementAssembler of unit "
				+ myUnit.getUnitNumber());

		    completed=true;
		    for (ii=0; ii<endUnitList.getCount(); ii++)
			if(quantitiesInEndUnits[ii]<0)completed=false;

		    if(completed)
			{
	if(StartVEFrame.verbose)
			    System.out.println("completed");
			    for (i=0;i<pendingProcurementSpecificationSet.
				     getNumberOfItemsToBeProcured();i++)
				for (ii=0; ii<endUnitList.getCount(); ii++)
				    if(endUnitLabels[ii]==
				       pendingProcurementSpecificationSet.
				       getItemToBeProcured(i))
					anOrder.setProcuredItemList((Order)
					((EndUnit) endUnitList.atOffset(ii)).
					    removeFromTemporaryList() );
			    
			    myUnit.setWaitingList(anOrder);
			}
		    else waitingList.addLast(anOrder);

		}
    }

    /**
     * return the waiting list length
     */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }


    /* PTHistogramPlottable interface method: getLabel() */
    public String getLabel(){
	return "";
    }

    /* PTHistogramPlottable interface method: getValueToPlot() */
    public double getValueToPlot(){
	return (double) getWaitingListLength();
    }

}



