// VEFrameObserverSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.simtoolsgui.GUISwarm;
import swarm.simtoolsgui.GUISwarmImpl;

import swarm.analysis.EZGraph;
import swarm.analysis.EZGraphImpl;

import java.util.*;


/**
 * The Observer contains the Model and the graphic widgets
 * @author  Pietro Terna and several of his students
 */
public class VEFrameObserverSwarm extends GUISwarmImpl {

    /** update frequency */
    public int displayFrequency;

    /** displaying all messages on the console */
    public boolean verboseChoice;

    /** two ActionGroup for sequence of GUI events */
    public ActionGroup displayActions1, displayActions2;

    /** the single Schedule instance */
    public Schedule displaySchedule;

    /** the Swarm we are observing */
    public VEFrameModelSwarm vEFrameModelSwarm;

    /** our graphics or EZGraph output to files*/
    public EZGraphImpl waitingListGraph, warehouseGraph, totalTimeLengthGraph,
	totalTimeLengthDataFile, totalDailyCostFile, totalCostFile,
        totalRevenueFile, dailyRevenueFile, dailySemimanufacturedOrderRevenueFile,
        totalInventoryFinancialCostFile, dailyStoredComponentValueFile,
        dailySemimanufacturedOrderCostFile, finishedOrderCostFile, benefitFile, 
	benefitGraph;

    /** a Java graphic via ptplot class*/
    public PTHistogram pTHistogram1, pTHistogram2;

    /** Constructor for class */
    public VEFrameObserverSwarm (Zone aZone) {

	super(aZone);

	// fill in the relevant parameters (only two, in this case).
	displayFrequency = 1;
	verboseChoice=false;

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  VEFrameObserverSwarm$1$VEFrameObserverProbeMap.class
	class VEFrameObserverProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, VEFrameObserverSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, VEFrameObserverSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }

	    public VEFrameObserverProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("displayFrequency");
		addVar("verboseChoice");
	    }
	}

	// Install our custom probeMap class directly into the
	// probeLibrary
	Globals.env.probeLibrary.setProbeMap$For
	    (new VEFrameObserverProbeMap (aZone, getClass ()), getClass ());
    }


    /** Create the objects used to display of the model. */
    public Object buildObjects () {

	super.buildObjects ();
	
	// First, we create the model that we're actually observing. The
	// model is a subswarm of the observer.

	// creating the VEFrameModelSwarm via lispAppArchivers, which
	// refers to the jveframe.scm file where we can modify parameters
	// in a stable way without compiling again the code
	vEFrameModelSwarm =
	    (VEFrameModelSwarm)Globals.env.lispAppArchiver.getWithZone$key(
						      getZone(), "vEFrameModelSwarm");
	// instead of using the direct constructor way
	// vEFrameModelSwarm = new VEFrameModelSwarm (getZone ());

	// Now create probe objects on the model and ourselves. This gives a
	// simple user interface to let the user change parameters.

	Globals.env.createArchivedProbeDisplay (vEFrameModelSwarm,
						"vEFrameModelSwarm");
	Globals.env.createArchivedProbeDisplay (this, "vEFrameObserverSwarm");

	// Instruct the control panel to wait for a button event: we
	// halt here until someone hits a control panel button so the
	// user can get a chance to fill in parameters before the
	// simulation runs
	getControlPanel ().setStateStopped ();

	// to change the verboseChoice via probe you have to wright 'true'
	// or 'false'; not simply 't' or 'f'

	// OK - the user has specified all the parameters for the
	// simulation.  Now we're ready to start.

	// setting the 'verbose' value in StartVEFrame
	StartVEFrame.verbose=verboseChoice;

	// First, let the model swarm build its objects.
	vEFrameModelSwarm.buildObjects ();

	// checking for logical consistency of dislayFrequency and 
	// ticksInADay
	if (! (1==displayFrequency || 0==displayFrequency%
	       vEFrameModelSwarm.getTicksInADay()) )
	    {
		System.out.println("Use displayFrequency == 1 or");
		System.out.println(
		"    displayFrequency == k*ticksInADay, with k integer.");
		System.exit(0);
	    }

	// Create the graph widget to display the waiting list content
	waitingListGraph = new EZGraphImpl
	    (getZone (), "Waiting lists", "Time",
	     "Waiting list values (avrg, min, max)","waiting_list");

	Globals.env.setWindowGeometryRecordName (waitingListGraph,
						 "waitingListGraph");

	// create the data for the waitingListGraph
	waitingListGraph.createAverageSequence$withFeedFrom$andSelector
	    ("avrgWaitingList", vEFrameModelSwarm.getUnitList(),
	     SwarmUtils.getSelector("Unit","getWaitingListLength"));
	waitingListGraph.createMinSequence$withFeedFrom$andSelector
	    ("minWaitingList", vEFrameModelSwarm.getUnitList(),
	     SwarmUtils.getSelector("Unit","getWaitingListLength"));

	waitingListGraph.createMaxSequence$withFeedFrom$andSelector
	    ("maxWaitingList", vEFrameModelSwarm.getUnitList(),
	     SwarmUtils.getSelector("Unit", "getWaitingListLength"));


	// Create the graph widget to display the quantities stored into 
	// the warehouse of each unit

	if (vEFrameModelSwarm.getUseWarehouses())
	    // if useWarehouses is true, draw the graph
	    {
		warehouseGraph = new EZGraphImpl
		    (getZone (), "Quantities in the warehouses", "Time",
		     "Quantities (avrg, min, max)", "warehouse_list");

		Globals.env.setWindowGeometryRecordName
		    (warehouseGraph, "warehouseGraph");

		// the data for the warehouseGraph
		warehouseGraph.createAverageSequence$withFeedFrom$andSelector
		    ("avrgQ", vEFrameModelSwarm.getWarehouseList(),
		     SwarmUtils.getSelector("Warehouse","getInventoryQuantity"));
		warehouseGraph.createMinSequence$withFeedFrom$andSelector
		    ("minQ", vEFrameModelSwarm.getWarehouseList(),
		     SwarmUtils.getSelector("Warehouse","getInventoryQuantity"));
		warehouseGraph.createMaxSequence$withFeedFrom$andSelector
		    ("maxQ", vEFrameModelSwarm.getWarehouseList(),
		     SwarmUtils.getSelector("Warehouse","getInventoryQuantity"));
	    }

	// Create the graph widget to display the total production time
	// for all the finished orders divided by their total recipe lengths

	totalTimeLengthGraph = new EZGraphImpl
	    (getZone (), "Ratio total time / total lengths", "Time",
	     "Ratio", "totalTimeLengthGraph");

	Globals.env.setWindowGeometryRecordName
	    (totalTimeLengthGraph, "totalTimeLengthGraph");

	// the data for the totalTimeLengthGraph
	totalTimeLengthGraph.createSequence$withFeedFrom$andSelector
	    ("ratio", vEFrameModelSwarm,
	     SwarmUtils.getSelector(vEFrameModelSwarm,"getTimeLengthRatio"));

	// the same, in a file ('data.ratio')
	totalTimeLengthDataFile =  new EZGraphImpl
	    (getZone(), "data");
	totalTimeLengthDataFile.createSequence$withFeedFrom$andSelector
	    ("ratio", vEFrameModelSwarm,
	     SwarmUtils.getSelector(vEFrameModelSwarm,"getTimeLengthRatio"));

	// Send the amount of totalDailyCosts of Units to the file 
	// totalDailyCosts.txt
	totalDailyCostFile =  new EZGraphImpl
	  (getZone(), "Costs/totalDailyCosts");
	totalDailyCostFile.createTotalSequence$withFeedFrom$andSelector
	    ("txt", vEFrameModelSwarm.getUnitList(),
	     SwarmUtils.getSelector("Unit","getTotalDailyCosts"));

	// Send the cumulated amount of fixed and variable costs of units 
	// from the beginning of the simulation to the file totalCosts.txt)
	totalCostFile =  new EZGraphImpl
	    (getZone(), "Costs/totalCosts");
	totalCostFile.createTotalSequence$withFeedFrom$andSelector
	    ("txt", vEFrameModelSwarm.getUnitList(),
	     SwarmUtils.getSelector("Unit","getTotalCosts"));

	// Send the revenues of the virtual enterprise to the file 
	// totalRevenues.txt
	totalRevenueFile =  new EZGraphImpl
	    (getZone(), "Revenues/totalRevenues");
	totalRevenueFile.createTotalSequence$withFeedFrom$andSelector
	    ("txt", vEFrameModelSwarm.getUnitList(),
	     SwarmUtils.getSelector("Unit","getTotalSaleRevenues"));

	// Send the daily revenue from the finished orders to the file
	// dailyRevenues
	dailyRevenueFile =  new EZGraphImpl
	    (getZone(), "Revenues/dailyRevenues");
	dailyRevenueFile.createTotalSequence$withFeedFrom$andSelector
	    ("txt", vEFrameModelSwarm.getUnitList(),
	     SwarmUtils.getSelector("Unit","getDailySaleRevenues"));

	// Send the revenues of the semimanufactured orders to the file
	// dailySemimanufacturedOrderRevenues.txt)
	dailySemimanufacturedOrderRevenueFile =  new EZGraphImpl
	    (getZone(), "Revenues/dailySemimanufacturedOrderRevenues");
	dailySemimanufacturedOrderRevenueFile.
	    createTotalSequence$withFeedFrom$andSelector
	    ("txt", vEFrameModelSwarm.getOrderList(),
	     SwarmUtils.getSelector("Order","getOrderEvaluation"));

	// Send the cumulated financial costs of the semi manufactured product 
	// in warehouses to the file totalInventoryFinancialCosts.txt
	if(vEFrameModelSwarm.getUseWarehouses()==true){
	    totalInventoryFinancialCostFile =  new EZGraphImpl
		(getZone(), "Costs/totalInventoryFinancialCosts");
	    totalInventoryFinancialCostFile.
		createTotalSequence$withFeedFrom$andSelector
		("txt", vEFrameModelSwarm.getUnitList(),
		 SwarmUtils.getSelector("Unit","getTotalInventoryFinancialCosts"));
	}

	// Send the revenue of the components stored in each day to the
	// file dailyStoredComponentValue
	if(vEFrameModelSwarm.getUseWarehouses()==true){
	    dailyStoredComponentValueFile =  new EZGraphImpl
		(getZone(), "Revenues/dailyStoredComponentValue");
	    dailyStoredComponentValueFile.
		createTotalSequence$withFeedFrom$andSelector
		("txt", vEFrameModelSwarm.getUnitList(),
		 SwarmUtils.getSelector("Unit","getDailyStoredComponentValue"));
	}


	// Send the costs of the orders made from the beginning of the 
	// simulation, to the file finishedOrderCosts.txt
	finishedOrderCostFile =  new EZGraphImpl
	    (getZone(), "Costs/finishedOrderCosts");
	finishedOrderCostFile.createTotalSequence$withFeedFrom$andSelector
	    ("txt", vEFrameModelSwarm.getUnitList(),
	     SwarmUtils.getSelector("Unit","getTotalFinishedOrderCosts"));

	// Send the daily cost of the orders in production to the
	// file dailySemimanufacturedOrderCosts.txt
	dailySemimanufacturedOrderCostFile =  new EZGraphImpl
	    (getZone(), "Costs/dailySemimanufacturedOrderCosts");
	dailySemimanufacturedOrderCostFile.
	    createTotalSequence$withFeedFrom$andSelector
	    ("txt", vEFrameModelSwarm.getOrderList(),
	     SwarmUtils.getSelector("Order","getOrderTotalCosts"));

	
	// the data for the benefitGraph
	benefitGraph = new EZGraphImpl
	    (getZone (), "Enterprise benefit", "Time",
	     "Benefit", "benefitGraph");
	Globals.env.setWindowGeometryRecordName
	    (benefitGraph, "benefitGraph");
	benefitGraph.createSequence$withFeedFrom$andSelector
	    ("benefitGraph", vEFrameModelSwarm,
	     SwarmUtils.getSelector(vEFrameModelSwarm,"getBenefit"));
	// the same in a file
	benefitFile =  new EZGraphImpl
	    (getZone(), "Benefit/benefit");
	benefitFile.createSequence$withFeedFrom$andSelector
	    ("txt", vEFrameModelSwarm,
	     SwarmUtils.getSelector(vEFrameModelSwarm,"getBenefit"));

	// the histogram for the waitingList of the units and for the
	// inventories in warehouses
	if(vEFrameModelSwarm.getUseWarehouses()==true){
	    pTHistogram1 = new PTHistogram(10,300,"Orders in Units", 
					   "Units", "Count", 
					   vEFrameModelSwarm.totalUnitNumber, 
					   20.0,
					   vEFrameModelSwarm.unitList,
					   "Queues in u.",
					   vEFrameModelSwarm.warehouseList,
					   "Quant. in w.",
					   vEFrameModelSwarm.
					   procurementAssemblerList,
					   "Q. in proc.ass.");
	}
        else pTHistogram1 = new PTHistogram(10,300,"Orders in Units", 
					    "Units", "Count", 
					    vEFrameModelSwarm.totalUnitNumber, 
					    20.0,
					    vEFrameModelSwarm.unitList,
					    "Queues in u.",
					    vEFrameModelSwarm.
					    procurementAssemblerList,
					    "Q. in proc.ass.");
	
	// the histogram for the waitingList of the end units
	if ( vEFrameModelSwarm.totalEndUnitNumber != 0)
	pTHistogram2 = new PTHistogram(10,250,
				       "Procurement (int. or ext.)", 
				       "End Units", "Count", 
				       vEFrameModelSwarm.totalEndUnitNumber, 
				       20.0,
				       vEFrameModelSwarm.endUnitList,
				       "q. in endU.");				       

	return this;
    }
    

  /**
   * Create the actions necessary for the simulation. This is where
   * the schedule is built (but not run!)  Here we create a display
   * schedule - this is used to display the state of the world and
   * check for user input. 
   */

    public Object buildActions () {
	int i;

	super.buildActions();

	// First, let our model swarm build its own schedule.
	vEFrameModelSwarm.buildActions();

	// Create two ActionGroup for display
	displayActions1 = new ActionGroupImpl (getZone());
	displayActions2 = new ActionGroupImpl (getZone());

	// Add the methods to the ActionGroup

	// Schedule the update of the graphic widgets

	displayActions1.createActionTo$message
	    (waitingListGraph,
	     SwarmUtils.getSelector(waitingListGraph, "step"));

	if(vEFrameModelSwarm.getUseWarehouses())
	    displayActions1.createActionTo$message(
       	       		  warehouseGraph,
	       		  SwarmUtils.getSelector(warehouseGraph,"step"));

	displayActions1.createActionTo$message
	    (totalTimeLengthGraph,
	     SwarmUtils.getSelector(totalTimeLengthGraph, "step"));

	displayActions1.createActionTo$message
	    (totalTimeLengthDataFile,
	     SwarmUtils.getSelector(totalTimeLengthGraph, "step"));

	// ptplot Java histograms
	displayActions1.createActionTo$message
	    (pTHistogram1,
	     SwarmUtils.getSelector(pTHistogram1, "addPoints"));
	if(vEFrameModelSwarm.totalEndUnitNumber != 0)
	displayActions1.createActionTo$message
	    (pTHistogram2,
	     SwarmUtils.getSelector(pTHistogram2, "addPoints"));

	// EZGraph files

	displayActions1.createActionTo$message
	    (totalDailyCostFile,
	     SwarmUtils.getSelector(totalDailyCostFile, "step"));

	displayActions1.createActionTo$message
	    (totalCostFile,
	     SwarmUtils.getSelector(totalCostFile, "step"));

	displayActions1.createActionTo$message
	    (totalRevenueFile,
	     SwarmUtils.getSelector(totalRevenueFile, "step"));

	displayActions1.createActionTo$message
	    (dailyRevenueFile,
	     SwarmUtils.getSelector(dailyRevenueFile, "step"));

	displayActions1.createActionTo$message
	    (dailySemimanufacturedOrderRevenueFile,
	     SwarmUtils.getSelector(dailySemimanufacturedOrderRevenueFile, "step"));

	if(vEFrameModelSwarm.getUseWarehouses()){
	    displayActions1.createActionTo$message
		(totalInventoryFinancialCostFile,
		 SwarmUtils.getSelector(totalInventoryFinancialCostFile, "step"));
	}

	if(vEFrameModelSwarm.getUseWarehouses()){
	    displayActions1.createActionTo$message
		(dailyStoredComponentValueFile,
		SwarmUtils.getSelector(dailyStoredComponentValueFile, "step"));
	}

	displayActions1.createActionTo$message
	    (dailySemimanufacturedOrderCostFile,
	     SwarmUtils.getSelector(dailySemimanufacturedOrderCostFile, "step"));

	displayActions1.createActionTo$message
	    (finishedOrderCostFile,
	     SwarmUtils.getSelector(finishedOrderCostFile, "step"));

	displayActions1.createActionTo$message
	    (benefitFile,
	     SwarmUtils.getSelector(benefitFile, "step"));

	displayActions1.createActionTo$message
	    (benefitGraph,
	     SwarmUtils.getSelector(benefitGraph, "step"));


	// Schedule the update of the probe displays
	displayActions2.createActionTo$message
	    (Globals.env.probeDisplayManager,
	     SwarmUtils.getSelector(Globals.env.probeDisplayManager,"update"));

	// Finally, schedule an update for the whole user
	// interface code.  This is crucial: without this, no
	// graphics update and the control panel will be
	// dead. It's best to put it at the end of the display
	// schedule
	displayActions2.createActionTo$message(getActionCache(),
       		      SwarmUtils.getSelector(getActionCache(),"doTkEvents"));

	// The display schedule. Note the repeat interval: display
	// is frequently the slowest part of a simulation, so redrawing less
	// frequently can be a help.
	displaySchedule = new ScheduleImpl (getZone (), displayFrequency);

	// insert ActionGroup instance on the repeating Schedule
	// instance

	for (i=0;i<displayFrequency;i++)
	    {
		if(i==displayFrequency-1)
		    displaySchedule.at$createAction (i, displayActions1);
		displaySchedule.at$createAction (i, displayActions2);
	    }
	
	return this;
    }

    /**
       activateIn: - activate the schedules so they're ready to run.
       The swarmContext argument has to do with what we were activated
       *in*.  Typically the ObserverSwarm is the top-level Swarm, so
       it's activated in "null". But other Swarms and Schedules and
       such will be activated inside of us.  */

    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves (just pass along the context).
	super.activateIn (swarmContext);

	// Activate the model swarm in ourselves. The model swarm is a
	// subswarm of the observer swarm.
	vEFrameModelSwarm.activateIn (this);

	// Now activate our schedule in ourselves. This arranges for
	// the execution of the schedule we built.
	displaySchedule.activateIn (this);

	// Activate returns the swarm activity - the thing that's ready to run.
	return getActivity();
  }

    /** drop the Observer, but after having dropped all the EZGraphs files,
	to close them */
    public void drop() {
	totalTimeLengthDataFile.drop();
	totalTimeLengthDataFile=null;

	totalDailyCostFile.drop();
	totalDailyCostFile=null;

	totalCostFile.drop();
	totalCostFile=null;

	totalRevenueFile.drop();
	totalRevenueFile=null;
	
	dailyRevenueFile.drop();
	dailyRevenueFile=null;
	
	dailySemimanufacturedOrderRevenueFile.drop();
	dailySemimanufacturedOrderRevenueFile=null;

	if(vEFrameModelSwarm.getUseWarehouses()==true){
	    totalInventoryFinancialCostFile.drop();
	    totalInventoryFinancialCostFile=null;
	}

	if(vEFrameModelSwarm.getUseWarehouses()==true){
	    dailyStoredComponentValueFile.drop();
	    dailyStoredComponentValueFile=null;
	}

	dailySemimanufacturedOrderCostFile.drop();
	dailySemimanufacturedOrderCostFile=null;

	finishedOrderCostFile.drop();
	finishedOrderCostFile=null;

	benefitFile.drop();
	benefitFile=null;

	super.drop();
    }


}











